/// Provides interoperability with [`url`](https://nodejs.org/api/url.html) module.
@JS()
library nodejs_interop.node.url;
// ignore_for_file: avoid_positional_boolean_parameters

import 'package:js/js.dart';

/// Provides options for serializing a URL object into a string.
@JS()
@anonymous
class FormatOptions {

  /// Creates a new options object.
  external factory FormatOptions({
    bool auth = true,
    bool fragment = true,
    bool search = true,
    bool unicode = false
  });

  /// Value indicating whether the serialized URL string should include the username and password.
  external bool get auth;
  external set auth(bool value);

  /// Value indicating whether the serialized URL string should include the fragment.
  external bool get fragment;
  external set fragment(bool value);

  /// Value indicating whether the serialized URL string should include the search query.
  external bool get search;
  external set search(bool value);

  /// Value indicating whether Unicode characters appearing in the host component of the URL string
  /// should be encoded directly as opposed to being [Punycode](https://tools.ietf.org/html/rfc3492) encoded.
  external bool get unicode;
  external set unicode(bool value);
}

/// Represents a URL.
@JS(r'$url.URL')
class Url {

  /// Creates a new URL, using the specified [base] URL to resolve against if the [input] is not absolute.
  external factory Url(String input, [base]);

  /// The fragment portion of the URL.
  external String get hash;
  external set hash(String value);

  /// The host portion of the URL.
  external String get host;
  external set host(String value);

  /// The hostname portion of the URL. The key difference between [host] and [hostname] is that [hostname] does not include the [port].
  external String get hostname;
  external set hostname(String value);

  /// The serialized URL.
  external String get href;
  external set href(String value);

  /// The read-only serialization of the URL's origin.
  external String get origin;

  /// The password portion of the URL.
  external String get password;
  external set password(String value);

  /// The path portion of the URL.
  external String get pathname;
  external set pathname(String value);

  /// The port portion of the URL.
  external int get port;
  external set port(int value);

  /// The protocol portion of the URL.
  external String get protocol;
  external set protocol(String value);

  /// The serialized query portion of the URL.
  external String get search;
  external set search(String value);

  /// The query parameters of the URL.
  external UrlSearchParams get searchParams;

  /// The username portion of the URL.
  external String get username;
  external set username(String value);

  /// Converts this object to a JSON representation.
  external String toJSON();
}

/// The [`url`](https://nodejs.org/api/url.html) module.
@JS()
abstract class UrlModule {

  /// Returns the [Punycode](https://tools.ietf.org/html/rfc3492) ASCII serialization of the [domain].
  /// If [domain] is an invalid domain, the empty string is returned.
  external String domainToASCII(String domain);

  /// Returns the Unicode serialization of the [domain].
  /// If [domain] is an invalid domain, the empty string is returned.
  external String domainToUnicode(String domain);

  /// Returns a formatted URL string derived from the specified [urlObject].
  external String format(urlObject, [FormatOptions options]);

  /// Takes a URL string, parses it, and returns a URL object.
  ///
  /// If [parseQueryString] is `true`, the query property will always be set to an object returned by the `QueryStringModule.parse()` method.
  /// If [slashesDenoteHost] is `true`, the first token after the literal string `//` and preceding the next `/` will be interpreted as the [UrlObject.host].
  external UrlObject parse(String urlString, [bool parseQueryString = false, bool slashesDenoteHost = false]);

  /// Resolves a target URL relative to a base URL in a manner similar to that of a Web browser resolving an anchor tag `href`.
  external String resolve(String from, String to);
}

/// An object whose properties represent significant elements of a URL.
@JS()
@anonymous
class UrlObject {

  /// Creates a new URL object.
  external factory UrlObject({
    String auth,
    String hash,
    String host,
    String hostname,
    String href,
    String path,
    String pathname,
    int port,
    String protocol,
    query,
    String search,
    bool slashes
  });

  /// The username and password portion of the URL, also referred to as "userinfo".
  external String get auth;
  external set auth(String value);

  /// The "fragment" portion of the URL including the leading hash (`#`) character.
  external String get hash;
  external set hash(String value);

  /// The full lower-cased host portion of the URL, including the [port] if specified.
  external String get host;
  external set host(String value);

  /// The lower-cased host name portion of the [host] component without the [port] included.
  external String get hostname;
  external set hostname(String value);

  /// The full URL string that was parsed with both the [protocol] and [host] components converted to lower-case.
  external String get href;
  external set href(String value);

  /// A concatenation of the [pathname] and [search] components.
  external String get path;
  external set path(String value);

  /// The entire path section of the URL.
  external String get pathname;
  external set pathname(String value);

  /// The numeric port portion of the [host] component.
  external int get port;
  external set port(int value);

  /// The URL's lower-cased protocol scheme.
  external String get protocol;
  external set protocol(String value);

  /// Either the query string without the leading question mark (`?`), or an object returned by the `QueryStringModule.parse()` method.
  external dynamic get query;
  external set query(value);

  /// The entire "query string" portion of the URL, including the leading question mark (`?`) character.
  external String get search;
  external set search(String value);

  /// Value indicating whether two forward-slash characters (`/`) are required following the colon in the [protocol].
  external bool get slashes;
  external set slashes(bool value);
}

/// Defines utility methods to work with the query string of a URL.
@JS(r'$url.URLSearchParams')
class UrlSearchParams {

  /// Creates new search parameters.
  external factory UrlSearchParams([query]);

  /// Appends a new [name]-[value] pair to the query string.
  external void append(String name, String value);

  /// Removes all name-value pairs whose name is [name].
  external void delete(String name);

  /// Returns an iterator over each of the name-value pairs in the query.
  external Iterable<List> entries();

  /// Iterates over each name-value pair in the query and invokes the given function.
  external void forEach(void fn(String value, String name, UrlSearchParams searchParams), [thisArg]);

  /// Returns the value of the first name-value pair whose name is [name]. If there are no such pairs, `null` is returned.
  external String get(String name);

  /// Returns the values of all name-value pairs whose name is [name]. If there are no such pairs, an empty list is returned.
  external List<String> getAll(String name);

  /// Returns a vale indicating whether there is at least one name-value pair whose name is [name].
  external bool has(String name);

  /// Returns an iterator over the names of each name-value pair.
  external Iterable<String> keys();

  /// Returns an iterator over the values of each name-value pair.
  external Iterable<String> values();

  /// Associates the specified [value] with the given [name].
  /// If there are any pre-existing name-value pairs whose names are [name], sets the first such pair's value to [value] and remove all others.
  external void set(String name, String value);

  /// Sorts all existing name-value pairs in-place by their names.
  external void sort();
}
