/// Provides interoperability with [`util`](https://nodejs.org/api/util.html) module.
@JS()
library nodejs_interop.node.util;

import 'package:js/js.dart';
import 'package:nodejs_interop/js.dart';

/// Interface defining a logging function.
typedef void LoggingFunction(data, [arg1, arg2, arg3, arg4, arg5, arg6, arg7]);

/// Options used to customize the text decoding.
@JS()
@anonymous
class DecodeOptions {

  /// Creates a new options object.
  external factory DecodeOptions({bool stream = false});

  /// Value indicating whether additional chunks of data are expected.
  external bool get stream;
  external set stream(bool value);
}

/// Inspects objects by returning a string representation that is primarily useful for debugging.
@JS()
abstract class InspectFunction implements Function {

  /// An object associating color names to their value.
  external dynamic get colors;

  /// A [`Symbol`](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Symbol)
  /// that can be used to declare custom `inspect` functions.
  external dynamic get custom;

  /// An object providing the default options used by this function.
  external InspectOptions get defaultOptions;

  /// An object associating style names to their color.
  external dynamic get styles;

  /// Returns a string representation of [object] that is primarily useful for debugging.
  /// Additional [options] may be passed that alter certain aspects of the formatted string.
  external String call(object, [InspectOptions options]);
}

/// Options used to alter certain aspects of a formatted string.
@JS()
@anonymous
class InspectOptions {

  /// Creates a new options object.
  external factory InspectOptions({
    int breakLength = 60,
    bool colors = false,
    bool customInspect = true,
    int depth = 2,
    int maxArrayLength = 100,
    bool showHidden = false,
    bool showProxy = false
  });

  /// The length at which an object's keys are split across multiple lines.
  /// Set to [double.INFINITY] to format an object as a single line.
  external int get breakLength;
  external set breakLength(int value);

  /// Value indicating whether the output will be styled with ANSI color codes.
  external bool get colors;
  external set colors(bool value);

  /// Value indicating whether custom inspection functions exported on the object being inspected will be called.
  external bool get customInspect;
  external set customInspect(bool value);

  /// Specifies the number of times to recurse while formatting the object.
  /// To make it recurse indefinitely pass `null`.
  external int get depth;
  external set depth(int value);

  /// Specifies the maximum number of list elements to include when formatting.
  /// Set to `null` to show all list elements. Set to `0` or negative to show no list elements.
  external int get maxArrayLength;
  external set maxArrayLength(int value);

  /// Value indicating whether the object's non-enumerable symbols and properties will be included in the formatted result.
  external bool get showHidden;
  external set showHidden(bool value);

  /// Value indicating whether objects and functions that are proxy objects will be introspected to show their target and handler objects.
  external bool get showProxy;
  external set showProxy(bool value);
}

/// Promisifies functions by returning a version that returns [Promise]s.
@JS()
abstract class PromisifyFunction implements Function {

  /// A [`Symbol`](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Symbol)
  /// that can be used to declare custom `promisify` functions.
  external dynamic get custom;

  /// Takes a [function] following the common Node.js callback style, i.e. taking a `(err, value) => ...` callback as the last argument,
  /// and returns a version that returns [Promise]s.
  external Function call(Function function);
}

/// Takes a stream of bytes as input and emits a stream of code points.
@JS(r'$util.TextDecoder')
class TextDecoder {

  /// Creates a new text decoder.
  external TextDecoder([String encoding, TextDecoderOptions options]);

  /// The encoding supported by this instance.
  external String get encoding;

  /// Value indicating whether decoding failures are fatal.
  external bool get fatal;

  /// Value indicating whether to include the byte order mark in the decoded result.
  external bool get ignoreBOM;

  /// Decodes the [input] array and returns a string containing the decoded bytes.
  external String decode([TypedArray input, DecodeOptions options]);
}

/// Options passed to text decoders.
@JS()
@anonymous
class TextDecoderOptions {

  /// Creates a new options object.
  external factory TextDecoderOptions({bool fatal = false, bool ignoreBOM = false});

  /// Value indicating whether decoding failures are fatal.
  external bool get fatal;
  external set fatal(bool value);

  /// Value indicating whether to include the byte order mark in the decoded result.
  /// This option is only used when encoding is `"utf-8"`, `"utf-16be"` or `"utf-16le"`.
  external bool get ignoreBOM;
  external set ignoreBOM(bool value);
}

/// Takes a stream of code points as input and emits a stream of bytes.
/// All instances of [TextEncoder] only support UTF-8 encoding.
@JS(r'$util.TextEncoder')
class TextEncoder {

  /// Creates a new text encoder.
  external TextEncoder();

  /// The encoding supported by this instance. Always set to `"utf-8"`.
  external String get encoding;

  /// Encodes the [input] string and returns a [Uint8Array] containing the encoded bytes.
  external Uint8Array encode([String input = '']);
}

/// The [`util`](https://nodejs.org/api/util.html) module.
@JS()
abstract class UtilModule {

  /// Inspects objects by returning a string representation that is primarily useful for debugging.
  external InspectFunction get inspect;

  /// Promisifies callback functions by returning a version that returns [Promise]s.
  external PromisifyFunction get promisify;

  /// Takes an `async` [function] (or a function that returns a [Promise])
  /// and returns a function following the Node.js error first callback style.
  ///
  /// In the callback, the first argument will be the rejection reason (or `null` if the promise resolved),
  /// and the second argument will be the resolved value.
  external Function callbackify(Function function);

  /// Creates a function that conditionally writes debug messages to `stderr` based on the existence of the `NODE_DEBUG` environment variable.
  /// If the [section] name appears within the value of that environment variable, then the returned function operates similar to `Console.error()`.
  /// If not, then the returned function is a no-op.
  external LoggingFunction debuglog(String section);

  /// Wraps the given [function] in such a way that it is marked as deprecated.
  /// Returns a function that will emit a `DeprecationWarning` using the `process.on("warning")` event.
  ///
  /// By default, the deprecation warning will be emitted and printed to `stderr` exactly once, the first time the returned function is called.
  /// After this warning is emitted, the wrapped function is called.
  external Function deprecate(Function function, [String message]);

  /// Returns a formatted string using the first argument as a `printf`-like format.
  ///
  /// The first argument is a string containing zero or more placeholder tokens.
  /// Each placeholder token is replaced with the converted value from the corresponding argument. Supported placeholders are:
  /// - `%s`: String.
  /// - `%d`: Number.
  /// - `%i`: Integer.
  /// - `%f`: Floating point value.
  /// - `%j`: JSON. Replaced with the string `"[Circular]"` if the argument contains circular references.
  /// - `%o`: Object. A string representation of an object with generic formatting. This will show the full object including non-enumerable symbols and properties.
  /// - `%O`: Object. A string representation of an object with generic formatting. This will show the full object excluding non-enumerable symbols and properties.
  /// - `%%`: Single percent sign (`"%"`). This does not consume an argument.
  external String format(String format, [arg1, arg2, arg3, arg4, arg5, arg6, arg7]);

  /// Inherits the prototype methods from one [constructor] into another.
  /// The prototype of [constructor] will be set to a new object created from [superConstructor].
  ///
  /// As an additional convenience, [superConstructor] will be accessible through the `constructor.super_` property.
  external void inherits(constructor, superConstructor);

  /// Returns a value indicating whether there is deep strict equality between [value1] and [value2].
  external bool isDeepStrictEqual(value1, value2);
}
